==> ota_tools_R35.1.0_aarch64.tbz2: Linux_for_Tegra/tools/ota_tools/version_upgrade/nv_ota_update_with_layout_change.sh <==
#!/bin/bash

# Copyright (c) 2019-2022, NVIDIA CORPORATION.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# This is a script to do preparation for OTA process

# 1. Validate base version and target board
# 2. Check key partitions (BCT/MB1/MB1_BCT) and avoid version rollback
# 3. Check whehter there is enough space for storing partitions of backup
#    boot path on the disk
# 4. Check partitons layout to avoid any partition in current system is
#    missing
# 5. Enable A/B redundancy if not enabled and check whether both slots are valid
# 6. Write base recovery image/dtb into "kernel"/"kernel-dtb" partitions
# 7. Set active to A
set -e

BASE_VERSION=
MSI_EMMC_MIN_SIZE=
MSI_EMMC_MIN_SIZE_TABLE=(
	# board:msi_emmc_min_size
	'jetson-tx2-devkit:134217728'    # 128MB
	'jetson-agx-xavier-devkit:335544320' # 320MB
	'jetson-xavier-nx-devkit-emmc:241172480' # 230MB
)
TARGET_BOARD=

usage()
{
	echo -ne "Usage: sudo $0 <disk for storing backup boot path>\n"
	echo -ne "\t<disk for storing backup boot path>:\n"
	echo -ne "\t    the device for storing the partitions as backup boot path.\n"
	echo -ne "Example:\n"
	echo -ne "\tsudo $0 /dev/mmcblk0\n"
	exit 1
}

both_slots_valid()
{
	# Check whether both slots are bootable
	# Usage:
	#        both_slots_valid
	local _nvbootctrl=
	_nvbootctrl="$(which nvbootctrl)"

	if [ "${_nvbootctrl}" = "" ]; then
		ota_log "nvbootctrl is not found"
		return 1
	fi

	local i=0
	while [ "$i" -lt 2 ]
	do
		if ! nvbootctrl is-slot-bootable $i; then
			ota_log "Slot $i is not bootable, device is not ready for OTA. Please do RCM flash to recover it"
			return 1
		fi
		if ! nvbootctrl is-slot-marked-successful $i; then
			ota_log "Slot $i is not marked successful, device is not ready for OTA. Please do RCM flash to recover it"
			return 1
		fi
		i=$((i + 1))
	done
	return 0
}

ota_check_free_space_on_emmc()
{
	# Get free space size on emmc
	# it's assumed that the free space can include 3 parts:
	# the first part is the real free space that is not allocated for any partition
	# the second part is the UDA partition that is not used in default case
	# the third part is the RECROOTFS partition that is not used in default case
	local Free_size=
	Free_size=$(parted /dev/mmcblk0 unit "B" print free | grep "Free Space" | tail -1 | sed 's/^[ \t]*//g' | sed 's/[ ][ ]*/ /g' | sed 's/B / /g' - | cut -d\  -f 3)
	if [ "${Free_size}" = "" ];then
		ota_log "Failed to get current free space on eMMC storage device"
		return 1
	fi
	local UDA_size=
	UDA_size=$(parted /dev/mmcblk0 unit "B" print | grep -m 1 "UDA" | sed 's/^[ ]*//g' | sed 's/[ ][ ]*/ /g' | cut -d\  -f 4 | sed 's/B$//g')
	if [ "${UDA_size}" = "" ];then
		ota_log "Failed to get size of UDA partition on eMMC storage device"
		return 1
	fi
	local RECROOTFS_size=
	RECROOTFS_size=$(parted /dev/mmcblk0 unit "B" print | grep -m 1 "RECROOTFS" | sed 's/^[ ]*//g' | sed 's/[ ][ ]*/ /g' | cut -d\  -f 4 | sed 's/B$//g')
	if [ "${RECROOTFS_size}" = "" ];then
		ota_log "Failed to get size of RECROOTFS partition on eMMC storage device"
		return 1
	fi

	local total_free_size=$((Free_size + UDA_size + RECROOTFS_size))
	if [ "${total_free_size}" -ge "${MSI_EMMC_MIN_SIZE}" ]; then
		ota_log "There is enough free space(${total_free_size} bytes > ${MSI_EMMC_MIN_SIZE} bytes) on eMMC"
		return 0
	else
		ota_log "Warning: no enough free space (${total_free_size} bytes < ${MSI_EMMC_MIN_SIZE} bytes) on eMMC"
		return 1
	fi
}

set_msi_emmc_min_size()
{
	# Set the MSI_EMMC_MIN_SIZE
	# Usage:
	#    set_msi_emmc_min_size {target_board} {_ret_msi_min_size}
	local target_board="${1}"
	local _ret_msi_min_size="${2}"
	local msi_min_size=

	local item=
	local board=
	local msi_min_size=
	for item in "${MSI_EMMC_MIN_SIZE_TABLE[@]}"
	do
		board="$(echo "${item}" | cut -d: -f 1)"
		if [ "${board}" == "${target_board}" ]; then
			msi_min_size="$(echo "${item}" | cut -d: -f 2)"
			eval "${_ret_msi_min_size}=${msi_min_size}"
			return 0
		fi
	done
	ota_log "Error: invalid target board(${target_board})"
	return 1
}

# Make sure that this script is running in root privilege
USERID=$(id -u)
if [ "${USERID}" -ne 0 ]; then
       echo "Please run this program as root."
       exit 0
fi

if [ $# -ne 1 ]; then
	usage
fi

nargs=$#;
OTA_DEVICE=${!nargs};

ota_log "Command: ${BASH_SOURCE[0]} $*"

if [ ! -e "${OTA_DEVICE}" ]; then
	echo "Invalid disk path ${OTA_DEVICE}"
	usage
fi

_parted="$(which parted)"
if [ "${_parted}" = "" ]; then
	echo "Please install the tool parted before running this script"
	exit 1
fi

_strings="$(which strings)"
if [ "${_strings}" = "" ]; then
	echo "Please install the tool strings (sudo apt-get install binutils) before running this script"
	exit
fi

_part_table="$(parted ${OTA_DEVICE} print | grep "Partition Table" | cut -d: -f 2 | sed 's/^[ ]\+//')"
if [ "${_part_table}" != "gpt" ]; then
	echo "Error: only gpt partition table is supported," \
		"please re-format disk with gpt partition table"
	exit 1
fi

# Handle the case of using SD card as backup boot path
# Comment it as it is outdated
: '
if [ "${OTA_DEVICE}" != "/dev/mmcblk0" ]; then
	OTA_WORK_DIR=/mnt/ota_work
	if [ ! -d "${OTA_WORK_DIR}" ]; then
		mkdir "${OTA_WORK_DIR}"
		chmod a+w "${OTA_WORK_DIR}"
	fi
	OTA_LOG_DIR=${OTA_PACKAGE_MOUNTPOINT}/ota_log
	source ./nv_ota_log.sh
	echo "init_ota_log ${OTA_LOG_DIR}"
	if ! init_ota_log "${OTA_LOG_DIR}"; then
		echo "Failed to run \"init_ota_log ${OTA_LOG_DIR}\""
		exit 1
	fi
	OTA_LOG_FILE="$(get_ota_log_file)"
	echo "OTA_LOG_FILE=${OTA_LOG_FILE}"
fi
'
source ./nv_ota_common_utils.func
source ./nv_ota_check_version.sh
#source ./ota_install_r32i_images_to_sd.sh
source ./ota_multi_board_specs.sh
source ./ota_check_partitions.sh
source ./nv_ota_common.func


# Check whether bsp version matches with OTA package version
ota_log "check_bsp_version ${OTA_WORK_DIR} ${OTA_DEVICE} BASE_VERSION"
if ! check_bsp_version "${OTA_WORK_DIR}" "${OTA_DEVICE}" "BASE_VERSION"; then
	ota_log "Failed to run \"check_bsp_version ${OTA_WORK_DIR} ${OTA_DEVICE} BASE_VERSION\""
	exit 1
fi

# Check whether current board matches with OTA package target board
ota_log "check_target_board ${OTA_WORK_DIR} TARGET_BOARD"
if ! check_target_board "${OTA_WORK_DIR}" "TARGET_BOARD"; then
	ota_log "Failed to run \"check_target_board ${OTA_WORK_DIR} TARGET_BOARD\""
	exit 1
fi

# Set MSI_EMMC_MIN_SIZE for corresponding board
ota_log "set_msi_emmc_min_size ${TARGET_BOARD} MSI_EMMC_MIN_SIZE"
if ! set_msi_emmc_min_size "${TARGET_BOARD}" "MSI_EMMC_MIN_SIZE"; then
	ota_log "Failed to run \"set_msi_emmc_min_size ${TARGET_BOARD} MSI_EMMC_MIN_SIZE\""
	exit 1
fi

# Check VER/VER_b partitions to avoid version rollback
ota_log "ota_check_rollback ${OTA_WORK_DIR} ${TARGET_BOARD} ${BASE_VERSION}"
if ! ota_check_rollback "${OTA_WORK_DIR}" "${TARGET_BOARD}" "${BASE_VERSION}"; then
	ota_log "Failed to run \"ota_check_rollback ${OTA_WORK_DIR} ${TARGET_BOARD} ${BASE_VERSION}\""
	exit 1
fi

# The board specific images are stored under images-XXX-XXX/<board_spec>/ directory.
# Need to parse current target board spec and then move the images from
# corresponding board_spec directory to image-XXX-XXX.
ota_log "ota_choose_images ${OTA_WORK_DIR} ${TARGET_BOARD} ${BASE_VERSION}"
if ! ota_choose_images "${OTA_WORK_DIR}" "${TARGET_BOARD}" "${BASE_VERSION}"; then
	ota_log "Failed to run \"ota_choose_images ${OTA_WORK_DIR} ${TARGET_BOARD} ${BASE_VERSION}\""
	exit 1
fi

if [ "${OTA_DEVICE}" != "/dev/mmcblk0" ];then
	ota_log "It's not supported to use non-onBoard-eMMC device as backup booting path"
	exit 1
#	ota_log "ota_install_r32i_images_to_sd ${OTA_WORK_DIR} ${OTA_DEVICE}"
#	if ! ota_install_r32i_images_to_sd "${OTA_WORK_DIR}" "${OTA_DEVICE}"; then
#		ota_log "Failed to run \"ota_install_r32i_images_to_sd ${OTA_WORK_DIR} ${OTA_DEVICE}\""
#		exit 1
#	fi
else
	# Check whether there is enough free space at the end of "${OTA_DEVICE}"
	# to store the MSI
	ota_log "ota_check_free_space_on_emmc"
	if ! ota_check_free_space_on_emmc; then
		ota_log "Failed to run \"ota_check_free_space_on_emmc\""
		exit 1
	fi

	# For upgrading R28.2 to R32 ToT, need to make sure that
	# APP partition has enough free space for adjustment.
	if [ "${BASE_VERSION}" == "R28-2" ];then
		source ./nv_ota_adjust_app.sh
		ota_log "ota_check_app_free_space ${OTA_WORK_DIR} /dev/root"
		if ! ota_check_app_free_space "${OTA_WORK_DIR}" "/dev/root"; then
			ota_log "Failed to run \"ota_check_app_free_space ${OTA_WORK_DIR} /dev/root\""
			exit 1
		fi
	fi

fi

# Check whether the offset of each partition in the "flash.idx" in the
# OTA payload package "ota_pakcage.tar.gz" matches the offset of each
# partition in current GPTs. For boot device, secondary GPT is used and
# for user device, primary GPT is used.
ota_log "ota_check_partitions ${OTA_WORK_DIR}"
if ! ota_check_partitions "${OTA_WORK_DIR}"; then
	ota_log "Failed to run \"ota_check_partitions ${OTA_WORK_DIR}\""
	exit 1
fi

# Enable A/B redundancy
ota_log "enable_a_b_redundancy"
if ! enable_a_b_redundancy; then
	ota_log "Failed to run \"enable_a_b_redundancy\""
	exit 1
fi

# Make sure both slots are bootable
ota_log "both_slots_valid"
if ! both_slots_valid; then
	ota_log "Failed to run \"both_slots_valid\""
	exit 1
fi

# R32.7 recovery image uses the new mechanism to handle the OTA process, so if base system is
# older than R32.7, need to write the r32.7 recovery image first.
if [ ! "${BASE_VERSION}" == "R32-7" ]; then
	ota_log "write_base_recovery ${OTA_WORK_DIR}"
	if ! write_base_recovery "${OTA_WORK_DIR}"; then
		ota_log "Failed to run \"write_base_recovery ${OTA_WORK_DIR}\""
		exit 1
	fi
fi

# For the platforms on which cboot loads kernel from filesystem, such as jetson-agx-xavier-devkit,
# the extlinux.conf needs to be renamed to force cboot load kernel from partition.
# Then the recovery image written in the kernel partition can be loaded.
ota_log "backup_extlinux_conf ${OTA_WORK_DIR}"
if ! backup_extlinux_conf "${OTA_WORK_DIR}"; then
	ota_log "Failed to run \"backup_extlinux_conf ${OTA_WORK_DIR}\""
	exit 1
fi

ota_log "write_kernel_bootctrl ${OTA_WORK_DIR}"
if ! write_kernel_bootctrl "${OTA_WORK_DIR}"; then
	ota_log "Failed to run \"write_kernel_bootctrl ${OTA_WORK_DIR}\""
	exit 1
fi

ota_log "OTA preprocess has been completed. OTA will be automatically started once device is rebooted"

==> ota_tools_r35.4.1_aarch64.tbz2: Linux_for_Tegra/tools/ota_tools/version_upgrade/nv_ota_update_with_layout_change.sh <==
#!/bin/bash

# Copyright (c) 2019-2023, NVIDIA CORPORATION.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# This is a script to do preparation for OTA process

# 1. Validate base version and target board
# 2. Check key partitions (BCT/MB1/MB1_BCT) and avoid version rollback
# 3. Check whehter there is enough space for storing partitions of backup
#    boot path on the disk
# 4. Check partitons layout to avoid any partition in current system is
#    missing
# 5. Enable A/B redundancy if not enabled and check whether both slots are valid
# 6. Write base recovery image/dtb into "kernel"/"kernel-dtb" partitions
# 7. Set active to A
set -e

BASE_VERSION=
MSI_EMMC_MIN_SIZE=
MSI_EMMC_MIN_SIZE_TABLE=(
	# board:msi_emmc_min_size
	'jetson-agx-xavier-devkit:335544320' # 320MB
	'jetson-agx-xavier-industrial:241172480' # 230MB
	'jetson-xavier-nx-devkit-emmc:241172480' # 230MB
)
NVME_USER_DEVICE=/dev/nvme0n1
SDMMC_USER_DEVICE=/dev/mmcblk0
TARGET_BOARD=

usage()
{
	echo -ne "Usage: sudo $0 <disk for storing backup boot path>\n"
	echo -ne "\t<disk for storing backup boot path>:\n"
	echo -ne "\t    the device for storing the partitions as backup boot path.\n"
	echo -ne "Example:\n"
	echo -ne "\tsudo $0 /dev/mmcblk0\n"
	exit 1
}

both_slots_valid()
{
	# Check whether both slots are bootable
	# Usage:
	#        both_slots_valid
	local _nvbootctrl=
	_nvbootctrl="$(which nvbootctrl)"

	if [ "${_nvbootctrl}" = "" ]; then
		ota_log "nvbootctrl is not found"
		return 1
	fi

	local i=0
	while [ "$i" -lt 2 ]
	do
		if ! nvbootctrl is-slot-bootable $i; then
			ota_log "Slot $i is not bootable, device is not ready for OTA. Please do RCM flash to recover it"
			return 1
		fi
		if ! nvbootctrl is-slot-marked-successful $i; then
			ota_log "Slot $i is not marked successful, device is not ready for OTA. Please do RCM flash to recover it"
			return 1
		fi
		i=$((i + 1))
	done
	return 0
}

ota_check_free_space_on_emmc()
{
	# Get free space size on emmc
	# it's assumed that the free space can include 3 parts:
	# the first part is the real free space that is not allocated for any partition
	# the second part is the UDA partition that is not used in default case
	# the third part is the RECROOTFS partition that is not used in default case
	local Free_size=
	Free_size=$(parted "${SDMMC_USER_DEVICE}" unit "B" print free | grep "Free Space" | tail -1 | sed 's/^[ \t]*//g' | sed 's/[ ][ ]*/ /g' | sed 's/B / /g' - | cut -d\  -f 3)
	if [ "${Free_size}" = "" ];then
		ota_log "Failed to get current free space on eMMC storage device"
		return 1
	fi
	local UDA_size=
	UDA_size=$(parted "${SDMMC_USER_DEVICE}" unit "B" print | grep -m 1 "UDA" | sed 's/^[ ]*//g' | sed 's/[ ][ ]*/ /g' | cut -d\  -f 4 | sed 's/B$//g')
	if [ "${UDA_size}" = "" ];then
		ota_log "Failed to get size of UDA partition on eMMC storage device"
		return 1
	fi
	local RECROOTFS_size=
	RECROOTFS_size=$(parted "${SDMMC_USER_DEVICE}" unit "B" print | grep -m 1 "RECROOTFS" | sed 's/^[ ]*//g' | sed 's/[ ][ ]*/ /g' | cut -d\  -f 4 | sed 's/B$//g')
	if [ "${RECROOTFS_size}" = "" ];then
		ota_log "Failed to get size of RECROOTFS partition on eMMC storage device"
		return 1
	fi

	local total_free_size=$((Free_size + UDA_size + RECROOTFS_size))
	if [ "${total_free_size}" -ge "${MSI_EMMC_MIN_SIZE}" ]; then
		ota_log "There is enough free space(${total_free_size} bytes > ${MSI_EMMC_MIN_SIZE} bytes) on eMMC"
		return 0
	else
		ota_log "Warning: no enough free space (${total_free_size} bytes < ${MSI_EMMC_MIN_SIZE} bytes) on eMMC"
		return 1
	fi
}

set_msi_emmc_min_size()
{
	# Set the MSI_EMMC_MIN_SIZE
	# Usage:
	#    set_msi_emmc_min_size {target_board} {_ret_msi_min_size}
	local target_board="${1}"
	local _ret_msi_min_size="${2}"
	local msi_min_size=

	local item=
	local board=
	local msi_min_size=
	for item in "${MSI_EMMC_MIN_SIZE_TABLE[@]}"
	do
		board="$(echo "${item}" | cut -d: -f 1)"
		if [ "${board}" == "${target_board}" ]; then
			msi_min_size="$(echo "${item}" | cut -d: -f 2)"
			eval "${_ret_msi_min_size}=${msi_min_size}"
			return 0
		fi
	done
	ota_log "Error: invalid target board(${target_board})"
	return 1
}

# Make sure that this script is running in root privilege
USERID=$(id -u)
if [ "${USERID}" -ne 0 ]; then
       echo "Please run this program as root."
       exit 0
fi

if [ $# -ne 1 ]; then
	usage
fi

nargs=$#;
OTA_DEVICE=${!nargs};

ota_log "Command: ${BASH_SOURCE[0]} $*"

if [ ! -e "${OTA_DEVICE}" ]; then
	echo "Invalid disk path ${OTA_DEVICE}"
	usage
fi

_parted="$(which parted)"
if [ "${_parted}" = "" ]; then
	echo "Please install the tool parted before running this script"
	exit 1
fi

_strings="$(which strings)"
if [ "${_strings}" = "" ]; then
	echo "Please install the tool strings (sudo apt-get install binutils) before running this script"
	exit
fi

_part_table="$(parted ${OTA_DEVICE} print | grep "Partition Table" | cut -d: -f 2 | sed 's/^[ ]\+//')"
if [ "${_part_table}" != "gpt" ]; then
	echo "Error: only gpt partition table is supported," \
		"please re-format disk with gpt partition table"
	exit 1
fi

source ./nv_ota_common_utils.func
source ./nv_ota_check_version.sh
source ./ota_multi_board_specs.sh
source ./ota_check_partitions.sh
source ./nv_ota_common.func

# Create the redundant copy of BCT in case it is not in place
if [ "${OTA_DEVICE}" != "${SDMMC_USER_DEVICE}" ]; then
	ota_log "duplicate_bct_copy"
	if ! duplicate_bct_copy; then
		ota_log "Failed to run \"dupclicate_bct_copy\""
		exit 1
	fi
fi

# Check whether bsp version matches with OTA package version
ota_log "check_bsp_version ${OTA_WORK_DIR} BASE_VERSION"
if ! check_bsp_version "${OTA_WORK_DIR}" "BASE_VERSION"; then
	ota_log "Failed to run \"check_bsp_version ${OTA_WORK_DIR} BASE_VERSION\""
	exit 1
fi

# Check whether current board matches with OTA package target board
ota_log "check_target_board ${OTA_WORK_DIR} TARGET_BOARD"
if ! check_target_board "${OTA_WORK_DIR}" "TARGET_BOARD"; then
	ota_log "Failed to run \"check_target_board ${OTA_WORK_DIR} TARGET_BOARD\""
	exit 1
fi

# Set MSI_EMMC_MIN_SIZE for corresponding board
ota_log "set_msi_emmc_min_size ${TARGET_BOARD} MSI_EMMC_MIN_SIZE"
if ! set_msi_emmc_min_size "${TARGET_BOARD}" "MSI_EMMC_MIN_SIZE"; then
	ota_log "Failed to run \"set_msi_emmc_min_size ${TARGET_BOARD} MSI_EMMC_MIN_SIZE\""
	exit 1
fi

# Check VER/VER_b partitions to avoid version rollback
ota_log "ota_check_rollback ${OTA_WORK_DIR} ${TARGET_BOARD} ${BASE_VERSION}"
if ! ota_check_rollback "${OTA_WORK_DIR}" "${TARGET_BOARD}" "${BASE_VERSION}"; then
	ota_log "Failed to run \"ota_check_rollback ${OTA_WORK_DIR} ${TARGET_BOARD} ${BASE_VERSION}\""
	exit 1
fi

# The board specific images are stored under images-XXX-XXX/<board_spec>/ directory.
# Need to parse current target board spec and then move the images from
# corresponding board_spec directory to image-XXX-XXX.
ota_log "ota_choose_images ${OTA_WORK_DIR}"
if ! ota_choose_images "${OTA_WORK_DIR}"; then
	ota_log "Failed to run \"ota_choose_images ${OTA_WORK_DIR}\""
	exit 1
fi

# Check whether there is enough free space at the end of internal eMMC
# to store the MSI
ota_log "ota_check_free_space_on_emmc"
if ! ota_check_free_space_on_emmc; then
	ota_log "Failed to run \"ota_check_free_space_on_emmc\""
	exit 1
fi

# Check whether the offset of each partition in the "flash.idx" in the
# OTA payload package "ota_pakcage.tar.gz" matches the offset of each
# partition in current GPTs. For boot device, secondary GPT is used and
# for user device, primary GPT is used.
ota_log "ota_check_partitions ${OTA_WORK_DIR}"
if ! ota_check_partitions "${OTA_WORK_DIR}"; then
	ota_log "Failed to run \"ota_check_partitions ${OTA_WORK_DIR}\""
	exit 1
fi

# Enable A/B redundancy
ota_log "enable_a_b_redundancy"
if ! enable_a_b_redundancy; then
	ota_log "Failed to run \"enable_a_b_redundancy\""
	exit 1
fi

# Make sure both slots are bootable
ota_log "both_slots_valid"
if ! both_slots_valid; then
	ota_log "Failed to run \"both_slots_valid\""
	exit 1
fi

# Write recovery and recovery-dtb partition
ota_log "write_base_recovery ${OTA_WORK_DIR}"
if ! write_base_recovery "${OTA_WORK_DIR}"; then
	ota_log "Failed to run \"write_base_recovery ${OTA_WORK_DIR}\""
	exit 1
fi

ota_log "write_kernel_bootctrl ${OTA_WORK_DIR}"
if ! write_kernel_bootctrl "${OTA_WORK_DIR}"; then
	ota_log "Failed to run \"write_kernel_bootctrl ${OTA_WORK_DIR}\""
	exit 1
fi

# Support NVMe device as external device
if [ "${OTA_DEVICE}" == "${NVME_USER_DEVICE}" ]; then
	# Force to boot from eMMC because the intermediate partition layout is on the eMMC device
	ota_log "force_booting_from_emmc ${OTA_WORK_DIR}"
	if ! force_booting_from_emmc "${OTA_WORK_DIR}"; then
		ota_log "Failed to run \"force_booting_from_emmc ${OTA_WORK_DIR}\""
		exit 1
	fi
fi

ota_log "OTA preprocess has been completed. OTA will be automatically started once device is rebooted"
